/**
 * class implementing fast waves uv
 */
object StellaExperiment2 {
  // define sample configuration
  val tiling0 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(256,256,64), OF)
  val tiling1 = Tiling(1, Memory.DeviceMemory, Memory.DeviceMemory, Domain(64,4,64), OF)
  val tiling2 = Tiling(2, Memory.RegisterFile, Memory.RegisterFile, Domain(Int.MaxValue,Int.MaxValue,1), HEKInc)
  val tiling3 = Tiling(3, Memory.ThreadMemory, Memory.SharedMemory, Domain(1,1,1), HEPar)

  val searchConfiguration = List((tiling3, true), (tiling2, false), (tiling1, true), (tiling0, true))

  // define stencils
  val ppgradcor1 = Stencil("ppgradcor1", 4, Map("wgtfac" -> AccessSet.center, "ppuv" -> AccessSet(List(Index(0,0,0), Index(0,0,-1)))))
  val ppgradcor2 = Stencil("ppgradcor2", 1, Map("ppgradcor1" -> AccessSet(List(Index(0,0,1), Index(0,0,0)))))
  val ppgradu = Stencil("ppgradu", 12, Map("ppuv" -> AccessSet(List(Index(0,0,0), Index(1,0,0))), "ppgradcor2" -> AccessSet(List(Index(0,0,0), Index(1,0,0))), "hhl" -> AccessSet(List(Index(0,0,0), Index(1,0,0), Index(0,0,1), Index(1,0,1)))))
  val ppgradv = Stencil("ppgradv", 12, Map("ppuv" -> AccessSet(List(Index(0,0,0), Index(0,1,0))), "ppgradcor2" -> AccessSet(List(Index(0,0,0), Index(0,1,0))), "hhl" -> AccessSet(List(Index(0,0,0), Index(0,1,0), Index(0,0,1), Index(0,1,1)))))
  val rhou = Stencil("rhou", 3, Map("rho" -> AccessSet(List(Index(0,0,0), Index(1,0,0)))))
  val rhov = Stencil("rhov", 3, Map("rho" -> AccessSet(List(Index(0,0,0), Index(0,1,0)))))
  val u1 = Stencil("u1", 4, Map("u0" -> AccessSet.center, "utens" -> AccessSet.center, "rhou" -> AccessSet.center, "ppgradu" -> AccessSet.center))
  val v1 = Stencil("v1", 4, Map("v0" -> AccessSet.center, "vtens" -> AccessSet.center, "rhov" -> AccessSet.center, "ppgradv" -> AccessSet.center))

  def runExperiment() = {
    println("Experiment2: FastWavesUV")
    println("========================")

    // evaluate the performance of variant 1
    {
      val group5 = StencilGroup(tiling3, List(ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"))
      val group4 = StencilGroup(tiling2, List(group5), List("u1", "v1"))

      val group3 = StencilGroup(tiling3, List(ppgradcor1, ppgradcor2), List("ppgradcor2"))
      val group2 = StencilGroup(tiling2, List(group3), List("ppgradcor2"))

      val group1 = StencilGroup(tiling1, List(group2, group4), List("u1", "v1"))
      val group0 = StencilGroup(tiling0, List(group1), List("u1", "v1"))

      println(s"-> evaluating storage requirements and performance")
      val t0 = System.nanoTime()
      val performance = group0.evaluatePerformance
      val requirements = group0.evaluateStorage
      val t1 = System.nanoTime()
      println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

      // evaluate the execution time for a specific solution
      println("\n1) Requirements (v1)")
      println("--------------------")
      println(requirements)
      println("-> " + requirements.verifyStorageRequirements)

      println("\n2) Performance (v1)")
      println("-------------------")

      println(performance.printNodeInformation("->"))
      println(performance.printAnalysisInformation("->"))
    }

    // evaluate the performance of variant 2
    {
      val group3 = StencilGroup(tiling3, List(ppgradcor1, ppgradcor2, ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"))
      val group2 = StencilGroup(tiling2, List(group3), List("u1", "v1"))
      val group1 = StencilGroup(tiling1, List(group2), List("u1", "v1"))
      val group0 = StencilGroup(tiling0, List(group1), List("u1", "v1"))

      println(s"-> evaluating storage requirements and performance")
      val t0 = System.nanoTime()
      val performance = group0.evaluatePerformance
      val requirements = group0.evaluateStorage
      val t1 = System.nanoTime()
      println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

      // evaluate the execution time for a specific solution
      println("\n1) Requirements (v2)")
      println("--------------------")
      println(requirements)
      println("-> " + requirements.verifyStorageRequirements)

      println("\n2) Performance (v2)")
      println("-------------------")

      println(performance.printNodeInformation("->"))
      println(performance.printAnalysisInformation("->"))
    }

    println("")
  }

  def runExperimentsGPU() = {
    // define sample configuration
    val tiling0 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(256,256,64), OF)
    val tiling1 = Tiling(1, Memory.DeviceMemory, Memory.DeviceMemory, Domain(64,4,64), OF)
    val tiling2 = Tiling(2, Memory.RegisterFile, Memory.RegisterFile, Domain(Int.MaxValue,Int.MaxValue,1), HEKInc)
    val tiling3 = Tiling(3, Memory.ThreadMemory, Memory.SharedMemory, Domain(1,1,1), HEPar)

    // no fusion
    {
      val group8 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, v1)), List("v1"))), List("v1"))
      val group7 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, u1)), List("u1"))), List("u1"))
      val group6 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, rhov)), List("rhov"))), List("rhov"))
      val group5 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, rhou)), List("rhou"))), List("rhou"))
      val group4 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, ppgradv)), List("ppgradv"))), List("ppgradv"))
      val group3 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, ppgradu)), List("ppgradu"))), List("ppgradu"))
      val group2 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, ppgradcor2)), List("ppgradcor2"))), List("ppgradcor2"))
      val group1 = StencilGroup(tiling1, List(StencilGroup(tiling2, List(StencilGroup(tiling3, ppgradcor1)), List("ppgradcor1"))), List("ppgradcor1"))
      val group0 = StencilGroup(tiling0, List(group1, group2, group3, group4, group5, group6, group7, group8), List("u1", "v1"))

      println(s"exp2 (GPU, no fusion): ${group0.executionTime} ${group0.isFeasible}")
    }

    // hand-tuned
    {
      val group5 = StencilGroup(tiling3, List(ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"))
      val group4 = StencilGroup(tiling2, List(group5), List("u1", "v1"))
      val group3 = StencilGroup(tiling3, List(ppgradcor1, ppgradcor2), List("ppgradcor2"))
      val group2 = StencilGroup(tiling2, List(group3), List("ppgradcor2"))
      val group1 = StencilGroup(tiling1, List(group2, group4), List("u1", "v1"))
      val group0 = StencilGroup(tiling0, List(group1), List("u1", "v1"))

      println(s"exp2 (GPU, hand-tuned): ${group0.executionTime} ${group0.isFeasible}")
    }

    // hand-tuned
    {
      val group3 = StencilGroup(tiling3, List(ppgradcor1, ppgradcor2, ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"))
      val group2 = StencilGroup(tiling2, List(group3), List("u1", "v1"))
      val group1 = StencilGroup(tiling1, List(group2), List("u1", "v1"))
      val group0 = StencilGroup(tiling0, List(group1), List("u1", "v1"))

      println(s"exp2 (GPU, optimized): ${group0.executionTime} ${group0.isFeasible}")
    }
  }

  def runExperimentsCPU() = {
    // define sample configuration
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(256, 256, 64), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostCache, Domain(8, 8, 64), OF)

    // no fusion
    {
      val group8 = StencilGroup(tiling1, v1)
      val group7 = StencilGroup(tiling1, u1)
      val group6 = StencilGroup(tiling1, rhov)
      val group5 = StencilGroup(tiling1, rhou)
      val group4 = StencilGroup(tiling1, ppgradv)
      val group3 = StencilGroup(tiling1, ppgradu)
      val group2 = StencilGroup(tiling1, ppgradcor2)
      val group1 = StencilGroup(tiling1, ppgradcor1)
      val group0 = StencilGroup(tiling0, List(group1, group2, group3, group4, group5, group6, group7, group8), List("u1", "v1"))

      println(s"exp2 (CPU, no fusion): ${group0.executionTime} ${group0.isFeasible}")
    }

    // hand-tuned
    {
      val group1 = StencilGroup(tiling1, List(ppgradcor1, ppgradcor2, ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"))
      val group0 = StencilGroup(tiling0, List(group1), List("u1", "v1"))

      println(s"exp2 (CPU, hand-tuned): ${group0.executionTime} ${group0.isFeasible}")
    }

    // optimized
    {
      val group1 = StencilGroup(tiling1, List(ppgradcor1, ppgradcor2, ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"))
      val group0 = StencilGroup(tiling0, List(group1), List("u1", "v1"))

      println(s"exp2 (CPU, optimized): ${group0.executionTime} ${group0.isFeasible}")
    }
  }

  def performOptimizationGPU(optimizationRange: (Int, Int)) = {
    println("Performing Optimization GPU")
    println("---------------------------")

    // define sample configuration
    val tiling0 = Tiling(0, Memory.DeviceMemory, Memory.DeviceMemory, Domain(256,256,64), OF)
    val tiling1 = Tiling(1, Memory.DeviceMemory, Memory.DeviceMemory, Domain(64,4,64), OF)
    val tiling2 = Tiling(2, Memory.RegisterFile, Memory.RegisterFile, Domain(Int.MaxValue,Int.MaxValue,1), HEKInc)
    val tiling3 = Tiling(3, Memory.ThreadMemory, Memory.SharedMemory, Domain(1,1,1), HEPar)

    val searchConfiguration = List((tiling3, true), (tiling2, false), (tiling1, true), (tiling0, true))

    val t0 = System.nanoTime()
    val optimizer = new OptimizerDP()
    val optimum = optimizer.searchOptimalTopology(optimizationRange, List(ppgradcor1, ppgradcor2, ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"), searchConfiguration)
    val t1 = System.nanoTime()

    println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

    val performance = optimum.evaluatePerformance
    val storage = optimum.evaluateStorage
    println(storage.prettyPrint("->"))
    println(performance.printNodeInformation("->"))
    println(performance.printAnalysisInformation("->"))

    println("")
  }

  def performOptimizationCPU(optimizationRange: (Int, Int)) = {
    println("Performing Optimization CPU")
    println("---------------------------")

    // search the optimal CPU solution
    val tiling0 = Tiling(0, Memory.HostMemory, Memory.HostMemory, Domain(256, 256, 64), OF)
    val tiling1 = Tiling(1, Memory.HostCache, Memory.HostCache, Domain(8, 8, 64), OF)
    val searchConfiguration = List((tiling1, true), (tiling0, true))

    val t0 = System.nanoTime()
    val optimizer = new OptimizerDP()
    val optimum = optimizer.searchOptimalTopology(optimizationRange, List(ppgradcor1, ppgradcor2, ppgradu, ppgradv, rhou, rhov, u1, v1), List("u1", "v1"), searchConfiguration)
    val t1 = System.nanoTime()

    println(s"-> elapsed time: ${(t1 - t0) / 1000 / 1000} ms")

    val performance = optimum.evaluatePerformance
    val storage = optimum.evaluateStorage
    println(storage.prettyPrint("->"))
    println(performance.printNodeInformation("->"))
    println(performance.printAnalysisInformation("->"))

    println("")
  }
}
