/**
 * class representing a single stencil
 */
case class Stencil(id: String, computeOperations: Long, dataDependencies: Map[String, AccessSet]) extends Node {
  val inputs = dataDependencies.toList.map(_._1)
  val outputs = List(id)
  val stencils = List(this)

  // compute two dependency maps considering the node dependencies and all nested stencil dependencies respectively
  val dependencies = {
    dataDependencies.map { dependency =>
      val relation = dependency._2.convertToISLMap
      val in = relation.setTupleId(isl.DimType.In, ISLContext.findOrCreateID(id))
      val out = in.setTupleId(isl.DimType.Out, ISLContext.findOrCreateID(dependency._1))
      isl.UnionMap.fromMap(out)
    }.reduce(_.union(_))
  }
  val flattenDependencies = dependencies

  // compute the stencil tiling by filtering the parent tiling
  def computeTilingImpl(parentOutputs: List[String], parentShifts: Map[String, List[Int]], parentTiling: isl.UnionMap, outputTiling: isl.UnionMap) = {
    // filter stencils / data fields that are not part of the group
    val space = isl.Space.setAlloc(ISLContext.ctx, 0, 3)
    val filter = isl.UnionSet.empty(space).addSet(isl.Set.universe(space).setTupleId(ISLContext.findOrCreateID(id)))
    List((parentTiling.intersectDomain(filter).coalesce(), outputTiling.coalesce()))
  }

  //override def toString = s"$id[Stencil]"
  override def toString = s"$id"

}
