import org.scalatest._

/**
 * test the cardinality computation
 */
class CardinalitySpec extends FlatSpec with Matchers {
  // these are our test sets
  val set1 = AccessSet.empty
  val set2 = AccessSet.center
  val set3 = AccessSet(List(Index(3,2,1), Index(-1,-2,-3)))
  val set4 = AccessSet(List(Index(2,3,-1), Index(0,0,0), Index(1,-2,3)))

  def simpleCardinality(domain: Domain, accesses: AccessSet) = {
    val interior = for {
      i <- 1 to domain.iSize
      j <- 1 to domain.jSize
      k <- 1 to domain.kSize
    } yield Index(i, j, k)
    (accesses + AccessSet(interior)).indexes.size
  }

  "a single element Domain" should "always evaluate to the access set size" in {
    val domain = Domain(1,1,1)
    set1.cardinality(domain) should be (set1.indexes.size)
    set2.cardinality(domain) should be (set2.indexes.size)
    set3.cardinality(domain) should be (set3.indexes.size)
    set4.cardinality(domain) should be (set4.indexes.size)
  }

  "an arbitrary Domain" should "always match the simple cardinality method" in {
    for(i <- 1 to 10; j <- 1 to 10; k <- 1 to 10) {
      val domain = Domain(i,j,k)
      withClue(s"set1 / iSize=$i, jSize=$j kSize=$k:") {
        set1.cardinality(domain) should be(simpleCardinality(domain, set1))
      }
      withClue(s"set2 / iSize=$i, jSize=$j kSize=$k:") {
        set2.cardinality(domain) should be(simpleCardinality(domain, set2))
      }
      withClue(s"set3 / iSize=$i, jSize=$j kSize=$k:") {
        set3.cardinality(domain) should be(simpleCardinality(domain, set3))
      }
      withClue(s"set4 / iSize=$i, jSize=$j kSize=$k:") {
        set4.cardinality(domain) should be(simpleCardinality(domain, set4))
      }
    }
  }
}
